"""
Implementations which handle various status codes.

Some operations include:
 * Re-authorising when a 'Not Authorised' status is returned.
 * Waiting until rate limit time has expired when a 'Too Many Requests' status is returned.
"""
import asyncio
import logging
from abc import ABC, abstractmethod
from collections.abc import Awaitable
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import NoReturn

from aiohttp import ClientResponse, ClientSession

from aiorequestful.auth import Authoriser
from aiorequestful.response.exception import ResponseError, StatusHandlerError
from aiorequestful.timer import Timer


class StatusHandler(ABC):
    """
    Handles a response that matches the status conditions of this handler
    according to its status, payload, headers, etc.
    """

    __slots__ = ("logger",)

    @property
    @abstractmethod
    def status_codes(self) -> list[int]:
        """The response status codes this handler can handle."""
        raise NotImplementedError

    def __init__(self):
        #: The :py:class:`logging.Logger` for this  object
        self.logger: logging.Logger = logging.getLogger(__name__)

    def match(self, response: ClientResponse, fail_on_error: bool = False) -> bool:
        """
        Check if this handler settings match the given response.

        :param response: The response to match.
        :param fail_on_error: Raise an exception if the response does not match this handler.
        :raise StatusHandlerError: If match fails and `fail_on_error` is True.
        """
        match = response.status in self.status_codes

        if not match and fail_on_error:
            raise StatusHandlerError(
                "Response status does not match this handler | "
                f"Response={response.status} | Valid={",".join(map(str, self.status_codes))}"
            )
        return match

    def _log(self, response: ClientResponse, message: str = ""):
        status = HTTPStatus(response.status)
        log = [f"Status: {response.status} ({status.phrase}) - {status.description}"]
        if message:
            log.append(message)

        self.logger.debug(" | ".join(log))

    @abstractmethod
    async def handle(self, response: ClientResponse, *_, **__) -> bool:
        """
        Handle the response based on its status, payload response etc.

        :param response: The response to handle.
        :return: True if the response was handled, False if it was not.
        """
        raise NotImplementedError

    def __call__(self, response: ClientResponse, *args, **kwargs) -> Awaitable[bool]:
        return self.handle(response, *args, **kwargs)


class ClientErrorStatusHandler(StatusHandler):
    """Handles status codes which cannot be handled, raising an error."""

    __slots__ = ()

    @property
    def status_codes(self) -> list[int]:
        return [status.value for status in HTTPStatus if 400 <= status.value < 500]

    async def handle(self, response: ClientResponse, *_, **__) -> NoReturn:
        self.match(response=response, fail_on_error=True)

        self._log(response=response, message="Bad response received and cannot handle or continue processing.")
        raise ResponseError(response=response)


class UnauthorisedStatusHandler(StatusHandler):
    """Handles unauthorised response status codes by re-authorising credentials through an :py:class:`Authoriser`."""

    __slots__ = ()

    @property
    def status_codes(self) -> list[int]:
        return [401]

    async def handle(
            self,
            response: ClientResponse,
            authoriser: Authoriser | None = None,
            session: ClientSession | None = None,
            *_,
            **__,
    ) -> bool:
        self.match(response=response, fail_on_error=True)
        if authoriser is None or session is None:
            return False

        self._log(response=response, message="Re-authorising...")
        headers = await authoriser
        session.headers.update(headers)
        return True


class RateLimitStatusHandler(StatusHandler):
    """Handles rate limits by increasing a timer value for every response that returns a rate limit status."""

    __slots__ = ("_wait_logged",)

    @property
    def status_codes(self) -> list[int]:
        return [429]

    def __init__(self):
        super().__init__()
        self._wait_logged = False

    async def handle(
            self,
            response: ClientResponse,
            wait_timer: Timer | None = None,
            retry_timer: Timer | None = None,
            *_,
            **__
    ) -> bool:
        self.match(response=response, fail_on_error=True)
        if wait_timer is not None:
            self._increase_wait(response=response, wait_timer=wait_timer)

        if "retry-after" not in response.headers:
            return False

        wait_seconds = int(response.headers["retry-after"])
        wait_dt_str = (datetime.now() + timedelta(seconds=wait_seconds)).strftime("%Y-%m-%d %H:%M:%S")

        if retry_timer is not None and wait_seconds > retry_timer.total:  # exception if too long
            raise ResponseError(
                "Rate limit exceeded and wait time is greater than remaining timeout "
                f"of {retry_timer.total_remaining:.2f} seconds. Retry again at {wait_dt_str}"
            )

        if not self._wait_logged:
            self.logger.warning(f"\33[93mRate limit exceeded. Retrying again at {wait_dt_str}\33[0m")
            self._wait_logged = True

        await asyncio.sleep(wait_seconds)
        self._wait_logged = False

        return True

    def _increase_wait(self, response: ClientResponse, wait_timer: Timer):
        if wait_timer.increase():
            self._log(
                response=response,
                message=f"Increasing wait time between requests to {float(wait_timer):.2f}s"
            )
        else:
            self._log(
                response=response,
                message=f"Cannot increase wait time. Already at maximum of {float(wait_timer):.2f}s"
            )



"""
Base interface for implementations of all cache backends.
"""
import asyncio
import logging
from abc import ABCMeta, abstractmethod
from collections.abc import MutableMapping, Callable, Collection, AsyncIterable, Mapping, Generator
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Self, Unpack

from aiohttp import RequestInfo, ClientRequest, ClientResponse
from dateutil.relativedelta import relativedelta

from aiorequestful._utils import get_iterator, classproperty
from aiorequestful.cache.exception import CacheError
from aiorequestful.response.exception import PayloadHandlerError
from aiorequestful.response.payload import PayloadHandler, StringPayloadHandler
from aiorequestful.types import UnitCollection, URLInput, RequestKwargs

type CacheRequestType = RequestInfo | ClientRequest | ClientResponse
type RepositoryRequestType[K] = K | CacheRequestType

DEFAULT_EXPIRE: timedelta = timedelta(weeks=1)


@dataclass
class ResponseRepositorySettings[V](metaclass=ABCMeta):
    """Settings for a response type from a given endpoint to be used to configure a repository in the cache backend."""
    #: That name of the repository in the backend
    name: str
    #: Handles payload data conversion to/from expected format for de/serialization.
    payload_handler: PayloadHandler[V] = field(default=StringPayloadHandler())

    @property
    @abstractmethod
    def fields(self) -> tuple[str, ...]:
        """
        The names of the fields relating to the keys extracted by :py:meth:`get_key` in the order
        in which they appear from the results of this method.
        """
        raise NotImplementedError

    @abstractmethod
    def get_key(self, **kwargs: Unpack[RequestKwargs]) -> tuple:
        """
        Extracts the name to assign to a cache entry in the repository from the given request kwargs.

        Arguments passed through to `.aiohttp.ClientSession.request`.
        See aiohttp reference for more info on available kwargs:
        https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientSession.request
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_name(payload: V) -> str | None:
        """Extracts the name to assign to a cache entry in the repository from a given ``response``."""
        raise NotImplementedError


class ResponseRepository[K: tuple, V: Any](AsyncIterable[tuple[K, V]], metaclass=ABCMeta):
    """
    Represents a repository in the backend cache, providing a dict-like interface
    for interacting with this repository.

    A repository is a data store within the backend e.g. a table in a database.

    :param settings: The settings to use to identify and interact with the repository in the backend.
    :param expire: The expiry time to apply to cached responses after which responses are invalidated.
    """

    __slots__ = ("logger", "connection", "settings", "_expire")

    # noinspection PyPropertyDefinition,PyMethodParameters
    @property
    @abstractmethod
    def _required_modules(cls) -> list:
        """The modules required to instantiate this repository"""
        return []

    @property
    def expire(self) -> datetime:
        """The datetime representing the maximum allowed expiry time from now."""
        return datetime.now() + self._expire

    @classmethod
    @abstractmethod
    def create(cls, *args, **kwargs) -> Self:
        """
        Set up the backend repository in the backend database if it doesn't already exist
        and return the initialised object that represents it.
        """
        raise NotImplementedError

    def __init__(self, settings: ResponseRepositorySettings[V], expire: timedelta | relativedelta = DEFAULT_EXPIRE):
        #: The :py:class:`logging.Logger` for this  object
        self.logger: logging.Logger = logging.getLogger(__name__)

        #: The settings to use to identify and interact with the repository in the backend.
        self.settings = settings
        self._expire = expire

        #: The current connection to the backend.
        self.connection = None

    @abstractmethod
    def __await__(self) -> Generator[None, None, Self]:
        raise NotImplementedError

    def __hash__(self):
        return hash(self.settings.name)

    @abstractmethod
    async def commit(self) -> None:
        """Commit the changes to the repository"""
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        """Close the connection to the repository."""
        raise NotImplementedError

    @abstractmethod
    async def count(self, include_expired: bool = True) -> int:
        """
        Get the number of responses in this repository.

        :param include_expired: Whether to include expired responses in the final count.
        :return: The number of responses in this repository.
        """
        raise NotImplementedError

    @abstractmethod
    async def contains(self, request: RepositoryRequestType[K]) -> bool:
        """Check whether the repository contains a given ``request``"""
        raise NotImplementedError

    @abstractmethod
    async def clear(self, expired_only: bool = False) -> int:
        """
        Clear the repository of all entries.

        :param expired_only: Whether to only remove responses that have expired.
        :return: The number of responses cleared from the repository.
        """
        raise NotImplementedError

    async def serialize(self, value: Any) -> V | None:
        """
        Serialize a given ``value`` to a type that can be persisted to the repository.

        :return: Serialized object if serializing is possible, None otherwise.
        """
        if value is None:
            return

        try:
            return await self.settings.payload_handler.serialize(value)
        except PayloadHandlerError:
            return

    async def deserialize(self, value: V | None) -> Any:
        """
        Deserialize a value from the repository to the expected response value type.

        :return: Deserialized object if deserializing is possible, None otherwise.
        """
        if value is None:
            return

        try:
            return await self.settings.payload_handler.deserialize(value)
        except PayloadHandlerError:
            return

    @abstractmethod
    def get_key_from_request(self, request: RepositoryRequestType[K]) -> K:
        """Extract the key to use when persisting responses for a given ``request``"""
        raise NotImplementedError

    @abstractmethod
    async def get_response(self, request: RepositoryRequestType[K]) -> V | None:
        """
        Get the response relating to the given ``request`` from this repository if it exists.

        :return: The result if found.
        """
        raise NotImplementedError

    async def get_responses(self, requests: Collection[RepositoryRequestType[K]]) -> list[V]:
        """
        Get the responses relating to the given ``requests`` from this repository if they exist.

        :return: Results unordered.
        """
        tasks = asyncio.gather(*map(self.get_response, requests))
        return list(filter(lambda result: result is not None, await tasks))

    async def save_response(self, response: Collection[K, V] | ClientResponse) -> None:
        """Save the given ``response`` to this repository if a key can be extracted from it. Safely fail if not"""
        if isinstance(response, Collection):
            key, value = response
        else:
            key = self.get_key_from_request(response)
            if not key:
                return

            value: V = await self.deserialize(response)

        await self._set_item_from_key_value_pair(key, await self.serialize(value))

    @abstractmethod
    async def _set_item_from_key_value_pair(self, __key: K, __value: Any) -> None:
        raise NotImplementedError

    async def save_responses(self, responses: Mapping[K, V] | Collection[ClientResponse]) -> None:
        """
        Save the given ``responses`` to this repository if a key can be extracted from them.
        Safely fail on those that can't.
        """
        if isinstance(responses, Mapping):
            tasks = [
                self._set_item_from_key_value_pair(key, await self.serialize(value))
                for key, value in responses.items()
            ]
        else:
            tasks = map(self.save_response, responses)

        await asyncio.gather(*tasks)

    @abstractmethod
    async def delete_response(self, request: RepositoryRequestType[K]) -> bool:
        """
        Delete the given ``request`` from this repository if it exists.

        :return: True if deleted in the repository and False if ``request`` was not found in the repository.
        """
        raise NotImplementedError

    async def delete_responses(self, requests: Collection[RepositoryRequestType[K]]) -> int:
        """
        Delete the given ``requests`` from this repository if they exist.

        :return: The number of the given ``requests`` deleted in the repository.
        """
        tasks = asyncio.gather(*map(self.delete_response, requests))
        return sum(await tasks)


class ResponseCache[R: ResponseRepository](MutableMapping[str, R], metaclass=ABCMeta):
    """
    Represents a backend cache of many repositories, providing a dict-like interface for interacting with them.

    :param cache_name: The name to give to this cache.
    :param repository_getter: A function that can be used to identify the repository in this cache
        that matches a given URL.
    :param expire: The expiry time to apply to cached responses after which responses are invalidated.
    """

    __slots__ = ("cache_name", "repository_getter", "expire", "_repositories")

    # noinspection PyMethodParameters
    @classproperty
    @abstractmethod
    def type(cls) -> str:
        """A string representing the type of the backend this class represents."""
        # raise NotImplementedError - omitted here as it causes docs build to fail

    @classmethod
    @abstractmethod
    async def connect(cls, value: Any, **kwargs) -> Self:
        """Connect to the backend from a given generic ``value``."""
        raise NotImplementedError

    def __init__(
            self,
            cache_name: str,
            repository_getter: Callable[[Self, URLInput], R | None] = None,
            expire: timedelta | relativedelta = DEFAULT_EXPIRE,
    ):
        super().__init__()

        #: The name to give to this cache.
        self.cache_name = cache_name
        #: A function that can be used to identify the repository in this cache that matches a given URL.
        self.repository_getter = repository_getter
        #: The expiry time to apply to cached responses after which responses are invalidated.
        self.expire = expire

        self._repositories: dict[str, R] = {}

    @abstractmethod
    def __await__(self) -> Generator[None, None, Self]:
        raise NotImplementedError

    async def __aenter__(self) -> Self:
        return await self

    @abstractmethod
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError

    def __repr__(self):
        return repr(self._repositories)

    def __str__(self):
        return str(self._repositories)

    def __iter__(self):
        return iter(self._repositories)

    def __len__(self):
        return len(self._repositories)

    def __getitem__(self, item):
        return self._repositories[item]

    def __setitem__(self, key, value):
        self._repositories[key] = value

    def __delitem__(self, key):
        del self._repositories[key]

    @abstractmethod
    async def commit(self) -> None:
        """Commit the changes to the cache"""
        raise NotImplementedError

    @abstractmethod
    async def close(self):
        """Close the connection to the repository."""
        raise NotImplementedError

    @abstractmethod
    def create_repository(self, settings: ResponseRepositorySettings) -> ResponseRepository:
        """
        Create and return a :py:class:`SQLiteResponseStorage` and store this object in this cache.

        Creates a repository with the given ``settings`` in the cache if it doesn't exist.
        """
        raise NotImplementedError

    def get_repository_from_url(self, url: URLInput) -> R | None:
        """Returns the repository to use from the stored repositories in this cache for the given ``url``."""
        if self.repository_getter is not None:
            return self.repository_getter(self, url)

    def get_repository_from_requests(self, requests: UnitCollection[CacheRequestType]) -> R | None:
        """Returns the repository to use from the stored repositories in this cache for the given ``requests``."""
        requests = get_iterator(requests)
        results = {self.get_repository_from_url(request.url) for request in requests}
        if len(results) > 1:
            raise CacheError(
                "Too many different types of requests given. Given requests must relate to the same repository type"
            )
        return next(iter(results), None)

    async def get_response(self, request: CacheRequestType) -> Any:
        """
        Get the response relating to the given ``request`` from the appropriate repository if it exists.

        :return: The result if found.
        """
        repository = self.get_repository_from_requests([request])
        if repository is not None:
            return await repository.get_response(request)

    async def get_responses(self, requests: Collection[CacheRequestType]) -> list:
        """
        Get the responses relating to the given ``requests`` from the appropriate repository if they exist.

        :return: Results unordered.
        """
        repository = self.get_repository_from_requests(requests)
        if repository is not None:
            return await repository.get_responses(requests)

    async def save_response(self, response: ClientResponse) -> None:
        """Save the given ``response`` to the appropriate repository if a key can be extracted from it."""
        repository = self.get_repository_from_requests([response])
        if repository is not None:
            return await repository.save_response(response)

    async def save_responses(self, responses: Collection[ClientResponse]) -> None:
        """
        Save the given ``responses`` to the appropriate repository if a key can be extracted from them.
        Safely fail on those that can't.
        """
        repository = self.get_repository_from_requests(responses)
        if repository is not None:
            return await repository.save_responses(responses)

    async def delete_response(self, request: CacheRequestType) -> bool:
        """
        Delete the given ``request`` from the appropriate repository if it exists.

        :return: True if deleted in the repository and False if ``request`` was not found in the repository.
        """
        repository = self.get_repository_from_requests([request])
        if repository is not None:
            return await repository.delete_response(request)

    async def delete_responses(self, requests: Collection[CacheRequestType]) -> int:
        """
        Delete the given ``requests`` from the appropriate repository.

        :return: The number of the given ``requests`` deleted in the repository.
        """
        repository = self.get_repository_from_requests(requests)
        if repository is not None:
            return await repository.delete_responses(requests)



"""
Implements basic authoriser flows.
"""
import base64

from aiorequestful.auth.base import Authoriser, _DEFAULT_SERVICE_NAME
from aiorequestful.types import Headers


class BasicAuthoriser(Authoriser):
    """
    Authorise HTTP requests using basic authentication i.e. username + password (optional)

    :param service_name: The service name for which to authorise.
    :param login: The login ID of the credentials.
    :param password: The login password.
    :param encoding: The encoding to apply to credentials when sending requests.
    """

    __slots__ = ("login", "password", "encoding")

    def __init__(
            self,
            login: str,
            password: str = "",
            encoding: str = "latin1",
            service_name: str = _DEFAULT_SERVICE_NAME
    ):
        super().__init__(service_name=service_name)

        #: The login ID of the credentials.
        self.login = login
        #: The login password.
        self.password = password
        #: The encoding to apply to credentials when sending requests.
        self.encoding = encoding

    async def authorise(self) -> Headers:
        credentials = f"{self.login}:{self.password}".encode(self.encoding)
        credentials_encoded = base64.b64encode(credentials).decode(self.encoding)
        return {
            "Authorization": f"Basic {credentials_encoded}"
        }


BASIC_CLASSES: frozenset[type[Authoriser]] = frozenset({
    BasicAuthoriser,
})



"""
Implements a ClientResponse to mock HTTP responses from cached data.
"""
from asyncio import StreamReader

from aiohttp import ClientRequest, ClientResponse
# noinspection PyProtectedMember
from aiohttp.helpers import TimerNoop
from multidict import CIMultiDictProxy


class CachedResponse(ClientResponse):
    """Emulates :py:class:`ClientResponse` for a response found in a cache backed."""

    def __init__(self, request: ClientRequest, payload: str | bytes):
        # noinspection PyTypeChecker,PyProtectedMember
        super().__init__(
            method=request.method,
            url=request.url,
            writer=None,
            continue100=None,
            timer=TimerNoop(),
            request_info=request.request_info,
            traces=[],
            loop=request.loop,
            session=request._session,
        )

        # response status
        self.version = request.version
        self.status = 200
        self.reason = "cached"

        # headers
        self._headers = CIMultiDictProxy(request.headers)
        self._raw_headers = ()

        self.content = StreamReader(loop=self._loop)

        if isinstance(payload, str):
            payload = payload.encode()
        self.content.feed_data(payload)
        self.content.feed_eof()



"""
All core type hints to use throughout the entire package.
"""
from __future__ import annotations

from collections.abc import Iterable, Sequence, MutableSequence, Collection, Mapping, MutableMapping, Callable, \
    Awaitable
from http import HTTPMethod
from types import SimpleNamespace
from typing import TypedDict, Any, Union, NotRequired

from aiohttp import BasicAuth, ClientResponse, ClientTimeout, Fingerprint
# noinspection PyProtectedMember
from aiohttp.client import SSLContext
# noinspection PyProtectedMember
from aiohttp.helpers import _SENTINEL
from aiohttp.typedefs import LooseCookies, LooseHeaders, StrOrURL
from yarl import URL

type UnitIterable[T] = T | Iterable[T]
type UnitCollection[T] = T | Collection[T]
type UnitSequence[T] = T | Sequence[T]
type UnitMutableSequence[T] = T | MutableSequence[T]
type UnitList[T] = T | list[T]

Number = int | float

ImmutableHeaders = Mapping[str, str]
MutableHeaders = MutableMapping[str, str]
Headers = dict[str, str]

JSON_VALUE = str | int | float | list | dict | bool | None
ImmutableJSON = Mapping[str, JSON_VALUE]
MutableJSON = MutableMapping[str, JSON_VALUE]
JSON = dict[str, JSON_VALUE]

URLInput = str | URL
MethodInput = str | HTTPMethod


class RequestKwargs(TypedDict):
    """
    TypedDict for kwargs relating to an HTTP request.

    Arguments passed through to `.aiohttp.ClientSession.request`.
    See aiohttp reference for more info on available kwargs:
    https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientSession.request
    """
    method: MethodInput
    url: URLInput
    params: NotRequired[Mapping[str, str]]
    data: NotRequired[Any]
    json: NotRequired[Any]
    cookies: NotRequired[LooseCookies]
    headers: NotRequired[LooseHeaders]
    skip_auto_headers: NotRequired[Iterable[str]]
    auth: NotRequired[BasicAuth]
    allow_redirects: NotRequired[bool]
    max_redirects: NotRequired[int]
    compress: NotRequired[str]
    chunked: NotRequired[bool]
    expect100: NotRequired[bool]
    raise_for_status: NotRequired[
        Union[bool, Callable[[ClientResponse], Awaitable[None]]]
    ]
    read_until_eof: NotRequired[bool]
    proxy: NotRequired[StrOrURL]
    proxy_auth: NotRequired[BasicAuth]
    timeout: NotRequired[Union[ClientTimeout, _SENTINEL]]
    verify_ssl: NotRequired[bool]
    fingerprint: NotRequired[bytes]
    ssl_context: NotRequired[SSLContext]
    ssl: NotRequired[Union[SSLContext, bool, Fingerprint]]
    server_hostname: NotRequired[str]
    proxy_headers: NotRequired[LooseHeaders]
    trace_request_ctx: NotRequired[SimpleNamespace]
    read_bufsize: NotRequired[int]
    auto_decompress: NotRequired[bool]
    max_line_size: NotRequired[int]
    max_field_size: NotRequired[int]



"""
Implementations to handle timers on the :py:class:`.RequestHandler`.

Includes various time incremental operations to manage increasing of time between requests
with various mathematical formulae.

Useful to handle backoff for requests on sensitive HTTP services
which often return a '429 - Too Many Requests' status.
"""
import asyncio
import functools
import itertools
from abc import ABC, ABCMeta, abstractmethod
from asyncio import sleep
from collections.abc import Generator
from copy import deepcopy
from typing import SupportsInt, SupportsFloat

from aiorequestful.types import Number


@functools.total_ordering
class Timer(SupportsInt, SupportsFloat, ABC):
    """
    Base interface for all timers.

    :param initial: The starting value to use.
    """

    __slots__ = ("_initial", "_value", "_counter")

    @property
    def initial(self) -> Number:
        """The initial starting timer value in seconds."""
        return self._initial

    @property
    @abstractmethod
    def final(self) -> Number | None:
        """The maximum possible timer value in seconds."""
        raise NotImplementedError

    @property
    @abstractmethod
    def total(self) -> Number | None:
        """The sum of all possible timer values in seconds."""
        raise NotImplementedError

    @property
    @abstractmethod
    def total_remaining(self) -> Number | None:
        """The sum of all possible remaining timer values in seconds not including the current value."""
        raise NotImplementedError

    @property
    @abstractmethod
    def count(self) -> int | None:
        """The total amount of times this timer can be increased."""
        raise NotImplementedError

    @property
    def counter(self) -> int | None:
        """The number of times this timer has been increased."""
        return self._counter

    @property
    def count_remaining(self) -> int | None:
        """The remaining number of times this timer can be increased."""
        if self.count is None:
            return None
        return self.count - self.counter

    @property
    @abstractmethod
    def can_increase(self) -> bool:
        """Check whether this timer can be increased"""
        raise NotImplementedError

    def __init__(self, initial: Number = 0):
        self._initial = initial
        self._value = initial
        self._counter = 0

    def __int__(self):
        return int(self._value)

    def __float__(self):
        return float(self._value)

    def __eq__(self, other: Number):
        return self._value == other

    def __lt__(self, other: Number):
        return self._value < other

    def __round__(self, n: int = None) -> float:
        return round(float(self._value), n)

    def __await__(self) -> Generator[None, None, None]:
        """Asynchronously sleep for the current time set for this timer."""
        return asyncio.sleep(self._value).__await__()

    def __call__(self) -> None:
        """Sleep for the current time set for this timer."""
        return self.wait()

    def __deepcopy__(self, memo: dict):
        cls = self.__class__
        obj = cls.__new__(cls)

        memo[id(self)] = obj
        slots = itertools.chain.from_iterable(getattr(c, '__slots__', []) for c in cls.__mro__)
        for key in slots:
            setattr(obj, key, deepcopy(getattr(self, key), memo))

        obj.reset()
        return obj

    @abstractmethod
    def increase(self) -> bool:
        """
        Increase the timer value.

        :return: True if timer was increased, False if not.
        """
        raise NotImplementedError

    def reset(self) -> None:
        """Reset the timer to its initial settings."""
        self._value = self._initial
        self._counter = 0

    def wait(self) -> None:
        """Sleep for the current time set for this timer."""
        sleep(self._value)


###########################################################################
## Count timers
###########################################################################
class CountTimer(Timer, metaclass=ABCMeta):
    """
    Abstract implementation of a :py:class:`Timer` which will increment a maximum number of times.

    :param initial: The starting value to use.
    :param count: The amount of times to increase the value.
    """

    __slots__ = ("_count",)

    @property
    def count(self):
        return self._count

    @property
    def can_increase(self) -> bool:
        return self.count is None or isinstance(self.count, int) and self.counter < self.count

    def __init__(self, initial: Number = 1, count: int = None):
        super().__init__(initial=initial)
        self._count = count


class StepCountTimer(CountTimer):
    """
    Increases timer value by a given ``step`` amount a distinct number of times.

    :param initial: The starting value to use.
    :param count: The amount of times to increase the value.
    :param step: The amount to increase the value by for each value increase.
    """

    __slots__ = ("_step",)

    @property
    def final(self):
        if self.count is None:
            return
        return self.initial + (self.step * self.count)

    @property
    def total(self):
        if self.count is None:
            return
        return sum(self.initial + self.step * i for i in range(self.count + 1))

    @property
    def total_remaining(self):
        if self.count is None:
            return
        return sum(float(self) + self.step * i for i in range(self.count_remaining + 1)) - float(self)

    @property
    def step(self) -> Number:
        """The amount to increase the timer value by in seconds."""
        return self._step

    def __init__(self, initial: Number = 0, count: int = None, step: Number = 1):
        super().__init__(initial=initial, count=count)
        self._step = step

    def increase(self) -> bool:
        if not self.can_increase:
            return False

        self._value += self.step
        self._counter += 1
        return True


class GeometricCountTimer(CountTimer):
    """
    Increases timer value by multiplying the current value by a given ``factor`` a distinct number of times.

    :param initial: The starting value to use.
    :param count: The amount of times to increase the value.
    :param factor: The amount to multiply the current value by for each value increase.
    """

    __slots__ = ("_factor",)

    @property
    def final(self):
        if self.count is None:
            return
        return self.initial * self.factor ** self.count

    @property
    def total(self):
        if self.count is None:
            return
        return sum(
            itertools.accumulate(range(1, self.count + 1), lambda s, _: s * self.factor, initial=self.initial)
        )

    @property
    def total_remaining(self):
        if self.count is None:
            return
        return sum(
            itertools.accumulate(range(self.count_remaining), lambda s, _: s * self.factor, initial=float(self))
        ) - float(self)

    @property
    def factor(self) -> Number:
        """The amount to multiply the timer value by in seconds."""
        return self._factor

    def __init__(self, initial: Number = 1, count: int = None, factor: Number = 2):
        super().__init__(initial=initial, count=count)
        self._factor = factor

    def increase(self) -> bool:
        if not self.can_increase:
            return False

        self._value *= self.factor
        self._counter += 1
        return True


class PowerCountTimer(CountTimer):
    """
    Increases timer value by raising the current value to a given ``exponent`` a distinct number of times.

    :param initial: The starting value to use.
    :param count: The amount of times to increase the value.
    :param exponent: The power to raise the value by for each value increase.
    """

    __slots__ = ("_exponent",)

    @property
    def final(self):
        if self.count is None:
            return
        return self.initial ** self.exponent ** self.count

    @property
    def total(self):
        if self.count is None:
            return
        return sum(
            itertools.accumulate(range(1, self.count + 1), lambda s, _: s ** self.exponent, initial=self.initial)
        )

    @property
    def total_remaining(self):
        if self.count is None:
            return
        return sum(
            itertools.accumulate(range(self.count_remaining), lambda s, _: s ** self.exponent, initial=float(self))
        ) - float(self)

    @property
    def exponent(self) -> Number:
        """The power value to apply to the timer value in seconds."""
        return self._exponent

    def __init__(self, initial: Number = 1, count: int = None, exponent: Number = 2):
        super().__init__(initial=initial, count=count)
        self._exponent = exponent

    def increase(self) -> bool:
        if not self.can_increase:
            return False

        self._value **= self.exponent
        self._counter += 1
        return True


###########################################################################
## Ceiling timers
###########################################################################
class CeilingTimer(Timer, metaclass=ABCMeta):
    """
    Abstract implementation of a :py:class:`Timer` which will increment until a maximum value is reached.

    :param initial: The starting value to use.
    :param final: The value at which to stop increasing.
    """

    __slots__ = ("_final",)

    @property
    def final(self):
        return self._final

    @property
    def total(self):
        if self.final is None:
            return
        return sum(self._all_values_iter(self.initial))

    @property
    def total_remaining(self):
        if self.final is None:
            return
        return sum(self._all_values_iter(float(self))) - float(self)

    @property
    def count(self):
        if self.final is None:
            return
        return len(list(self._all_values_iter(self.initial))) - 1

    @property
    def can_increase(self) -> bool:
        return self.final is None or isinstance(self.final, Number) and self < self.final

    def __init__(self, initial: Number = 1, final: Number = None):
        super().__init__(initial=initial)

        if final is not None and final < initial:
            final = initial

        self._final = final

    def _all_values_iter(self, value: Number) -> Generator[Number, None, None]:
        """Returns an iterator for all values remaining from the given ``value`` including the given ``value``."""
        raise NotImplementedError


class StepCeilingTimer(CeilingTimer):
    """
    Increases timer value by a given ``step`` amount until a maximum value is reached.

    :param initial: The starting value to use.
    :param final: The value at which to stop increasing.
    :param step: The amount to increase the value by for each value increase.
    """

    __slots__ = ("_step",)

    @property
    def step(self) -> Number:
        """The amount to increase the timer value by in seconds."""
        return self._step

    def __init__(self, initial: Number = 1, final: Number = None, step: Number = 1):
        super().__init__(initial=initial, final=final)
        self._step = step

    def _all_values_iter(self, value: Number) -> Generator[Number, None, None]:
        if self.final is None:
            return

        yield value

        while value < self.final:
            value = min(self.final, value + self.step)
            yield value

    def increase(self) -> bool:
        if not self.can_increase:
            return False

        if self.final is not None:
            self._value = min(self.final, self._value + self.step)
        else:
            self._value += self.step

        self._counter += 1
        return True


class GeometricCeilingTimer(CeilingTimer):
    """
    Increases timer value by multiplying the current value by a given ``factor`` until a maximum value is reached.

    :param initial: The starting value to use.
    :param final: The value at which to stop increasing.
    :param factor: The amount to multiply the current value by for each value increase.
    """

    __slots__ = ("_factor",)

    @property
    def factor(self) -> Number:
        """The amount to multiply the timer value by in seconds."""
        return self._factor

    def __init__(self, initial: Number = 1, final: Number = None, factor: Number = 2):
        super().__init__(initial=initial, final=final)
        self._factor = factor

    def _all_values_iter(self, value: Number) -> Generator[Number, None, None]:
        if self.final is None:
            return

        yield value

        while value < self.final:
            value = min(self.final, value * self.factor)
            yield value

    def increase(self) -> bool:
        if not self.can_increase:
            return False

        if self.final is not None:
            self._value = min(self.final, self._value * self.factor)
        else:
            self._value *= self.factor

        self._counter += 1
        return True


class PowerCeilingTimer(CeilingTimer):
    """
    Increases timer value by raising the current value to a given ``exponent`` until a maximum value is reached.

    :param initial: The starting value to use.
    :param final: The value at which to stop increasing.
    :param exponent: The power to raise the value by for each value increase.
    """

    __slots__ = ("_exponent",)

    @property
    def exponent(self) -> Number:
        """The power value to apply to the timer value in seconds."""
        return self._exponent

    def __init__(self, initial: Number = 1, final: Number = None, exponent: Number = 2):
        super().__init__(initial=initial, final=final)
        self._exponent = exponent

    def _all_values_iter(self, value: Number) -> Generator[Number, None, None]:
        if self.final is None:
            return

        yield value

        while value < self.final:
            value = min(self.final, value ** self.exponent)
            yield value

    def increase(self) -> bool:
        if not self.can_increase:
            return False

        if self.final is not None:
            self._value = min(self.final, self._value ** self.exponent)
        else:
            self._value **= self.exponent

        self._counter += 1
        return True



"""
Resources relating handling of responses from HTTP requests.

Includes operations to:
 * validate the success of HTTP requests
 * handle unsuccessful HTTP requests
 * deserialize response payload data to Python objects
"""



"""Welcome to aiorequestful"""
from pathlib import Path

PROGRAM_NAME = "aiorequestful"
PROGRAM_OWNER_NAME = "George Martin Marino"
PROGRAM_OWNER_USER = "geo-martino"
PROGRAM_OWNER_EMAIL = f"gm.engineer+{PROGRAM_NAME.lower()}@pm.me"
PROGRAM_URL = f"https://github.com/{PROGRAM_OWNER_USER}/{PROGRAM_NAME.lower()}"

MODULE_ROOT: str = Path(__file__).parent.name
PACKAGE_ROOT: Path = Path(__file__).parent.parent



"""
Core exceptions for the entire package.
"""


class AIORequestfulError(Exception):
    """Generic base class for all aiorequestful-related errors"""


class InputError(AIORequestfulError, ValueError):
    """Exception raised when the given input is invalid."""


class AIORequestfulImportError(AIORequestfulError, ImportError):
    """Exception raised for import errors, usually from missing modules."""


class HTTPError(AIORequestfulError):
    """Exception raised for generic HTTP errors."""


class RequestError(HTTPError):
    """Exception raised for errors relating to HTTP requests."""



"""
All operations relating to handling of requests to an HTTP service.

This is the core module for handling requests and their responses
including using authorisation and caching as necessary.
"""
from __future__ import annotations

import contextlib
import inspect
import json
import logging
from collections.abc import Mapping, Callable, Sequence
from copy import deepcopy
from http import HTTPMethod
from typing import Any, Self, Unpack
from urllib.parse import unquote

import aiohttp
from yarl import URL

from aiorequestful._utils import format_url_log
from aiorequestful.auth import Authoriser
from aiorequestful.cache.backend import ResponseCache
from aiorequestful.cache.session import CachedSession
from aiorequestful.exception import RequestError
from aiorequestful.response.payload import StringPayloadHandler, PayloadHandler
from aiorequestful.response.status import StatusHandler, ClientErrorStatusHandler, UnauthorisedStatusHandler, \
    RateLimitStatusHandler
from aiorequestful.timer import Timer
from aiorequestful.types import URLInput, Headers, MethodInput, RequestKwargs

_DEFAULT_RESPONSE_HANDLERS = [
    UnauthorisedStatusHandler(), RateLimitStatusHandler(), ClientErrorStatusHandler()
]


class RequestHandler[A: Authoriser, P: Any]:
    """
    Generic HTTP request handler.
    Handles error responses, retries on failed requests, authorisation, caching etc.

    :param connector: When called, returns a new session to use when making requests.
    :param authoriser: The authoriser to use when authorising requests to the HTTP service.
    :param payload_handler: Handles payload data conversion to return response payload in expected format.
    :param response_handlers: Handlers to handle responses for specific status codes.
        Should many of the given handlers be responsible for handling the same status code,
        the first handler in the sequence will be used for that status code.
    :param wait_timer: The time to wait after every request, regardless of whether it was successful.
        It is best practice to configure this such that a maximum time can be achieved
        within a reasonable time to ensure times between requests do not get too large.
        Useful to help manage calls to services that have strict restraints around rate limiting.
    :param retry_timer: The timer that controls how long to wait in between each successive failed request.
        It is best practice to configure this such that a maximum time can be achieved
        within a reasonable time to cause a timeout and raise an exception.
    """

    __slots__ = (
        "logger",
        "_connector",
        "_session",
        "authoriser",
        "_payload_handler",
        "_response_handlers",
        "wait_timer",
        "_retry_timer",
        "_retry_logged",
    )

    @property
    def closed(self):
        """Is the stored client session closed."""
        return self._session is None or self._session.closed

    @property
    def session(self) -> aiohttp.ClientSession:
        """The :py:class:`ClientSession` object if it exists and is open."""
        if not self.closed:
            return self._session

    @property
    def payload_handler(self) -> PayloadHandler:
        """Handles a response according to its status, payload, headers, etc."""
        return self._payload_handler

    @payload_handler.setter
    def payload_handler(self, value: PayloadHandler):
        self._payload_handler = value

        if isinstance(self._session, CachedSession):
            for repository in self._session.cache.values():
                # all repositories must use the same payload handler as the request handler
                # for it to function correctly
                repository.settings.payload_handler = self._payload_handler

    @property
    def response_handlers(self) -> dict[int, StatusHandler]:
        """Handles a response according to its status, payload, headers, etc."""
        return self._response_handlers

    @response_handlers.setter
    def response_handlers(self, value: Sequence[StatusHandler]):
        self._response_handlers.update({
            status: handler for handler in reversed(value) for status in handler.status_codes
        })

    @property
    def retry_timer(self) -> Timer | None:
        """
        The timer that controls how long to wait in between each successive failed request.
        Always returns a reset timer with initial settings.
        """
        if not self._retry_timer:
            return
        return deepcopy(self._retry_timer)

    @retry_timer.setter
    def retry_timer(self, value: Timer | None):
        self._retry_timer = value

    @classmethod
    def create(
            cls,
            authoriser: A | None = None,
            cache: ResponseCache | None = None,
            payload_handler: PayloadHandler[P] = None,
            response_handlers: Sequence[StatusHandler] = (),
            wait_timer: Timer = None,
            retry_timer: Timer = None,
            **session_kwargs
    ) -> RequestHandler[A, P]:
        """Create a new :py:class:`RequestHandler` with an appropriate session ``connector`` given the input kwargs"""
        def connector() -> aiohttp.ClientSession:
            """Create an appropriate session ``connector`` given the input kwargs"""
            if cache is not None:
                return CachedSession(cache=cache, **session_kwargs)
            return aiohttp.ClientSession(**session_kwargs)

        return cls(
            connector=connector,
            authoriser=authoriser,
            payload_handler=payload_handler,
            response_handlers=response_handlers,
            wait_timer=wait_timer,
            retry_timer=retry_timer,
        )

    def __init__(
            self,
            connector: Callable[[], aiohttp.ClientSession],
            authoriser: A | None = None,
            payload_handler: PayloadHandler = None,
            response_handlers: Sequence[StatusHandler] = (),
            wait_timer: Timer = None,
            retry_timer: Timer = None,
    ):
        #: The :py:class:`logging.Logger` for this  object
        self.logger: logging.Logger = logging.getLogger(__name__)

        self._connector = connector
        self._session: aiohttp.ClientSession | CachedSession | None = None

        #: The :py:class:`Authoriser` object
        self.authoriser = authoriser

        #: Handles payload data conversion to return response payload in expected format
        self.payload_handler = payload_handler if payload_handler is not None else StringPayloadHandler()

        self._response_handlers = {}
        self.response_handlers = response_handlers if response_handlers is not None else _DEFAULT_RESPONSE_HANDLERS

        #: The time to wait after every request, regardless of whether it was successful
        self.wait_timer = wait_timer
        self._retry_timer = retry_timer

        self._retry_logged = False

    async def __aenter__(self) -> Self:
        self._retry_logged = False

        if self.closed:
            self._session = self._connector()

        # force setting payload handler on all cache repositories
        self.payload_handler = self.payload_handler

        await self.session.__aenter__()
        await self.authorise()

        return self

    async def __aexit__(self, __exc_type, __exc_value, __traceback) -> None:
        if self._session is not None:
            await self._session.__aexit__(__exc_type, __exc_value, __traceback)
            self._session = None

        self._retry_logged = False

    async def authorise(self) -> Headers:
        """
        Authenticate and authorise, testing/refreshing/re-authorising as needed.
        Updates the session with new credentials.

        :return: Headers for request authorisation.
        :raise AuthoriserError: If the token cannot be validated.
        """
        if self.closed:
            raise RequestError("Session is closed. Enter this object's context to start a new session.")

        headers = {}
        if self.authoriser is not None:
            self.session.headers.update(await self.authoriser)

        return headers

    async def close(self) -> None:
        """Close the current session. No more requests will be possible once this has been called."""
        await self.session.close()

    def log(
            self, method: str, url: URLInput, message: str | list = None, level: int = logging.DEBUG, **kwargs
    ) -> None:
        """Format and log a request or request adjacent message to the given ``level``."""
        log: list[Any] = []

        url = URL(url)
        if url.query:
            log.extend(f"{k}: {unquote(v):<4}" for k, v in sorted(url.query.items()))
        if kwargs.get("params"):
            log.extend(f"{k}: {v:<4}" for k, v in sorted(kwargs.pop("params").items()))
        if kwargs.get("json"):
            log.extend(f"{k}: {str(v):<4}" for k, v in sorted(kwargs.pop("json").items()))
        if len(kwargs) > 0:
            log.extend(f"{k.title()}: {str(v):<4}" for k, v in kwargs.items() if v)
        if message:
            log.append(message) if isinstance(message, str) else log.extend(message)

        self.logger.log(level=level, msg=format_url_log(method=method, url=url, messages=log))

    async def request(self, method: MethodInput, url: URLInput, **kwargs: Unpack[RequestKwargs]) -> P:
        """
        Generic method for handling HTTP requests handling errors, authorisation, backoff, caching etc. as configured.
        See :py:func:`request` for more arguments.

        :param method: HTTP request method (such as GET, POST, PUT, etc.)
        :param url: The URL to perform the request on.
        :return: The JSON formatted response or, if JSON formatting not possible, the text response.
        :raise RequestError: For any request which fails.
        :raise ResponseError: For any request which returns an invalid response.
        :raise StatusHandlerError: For any request which returns a response with a status that could not be handled.
        """
        if self.closed:
            raise RequestError(
                "Could not send a request as the session is closed. "
                "Enter the RequestHandler's context to start a new session."
            )

        method = HTTPMethod(method.upper())
        retry_timer = self.retry_timer

        while True:
            async with self._request(method=method, url=url, **kwargs) as response:
                if response is None:
                    await self._handle_retry_timer(method=method, url=url, timer=retry_timer)
                    continue

                if await self._handle_response(response, retry_timer=retry_timer):
                    continue

                if response.ok:
                    payload: P = await self.payload_handler(response)
                    break

                await self._log_response(response=response, method=method, url=url)
                await self._handle_retry_timer(method=method, url=url, timer=retry_timer)

        self._retry_logged = False
        return payload

    @contextlib.asynccontextmanager
    async def _request(
            self,
            method: HTTPMethod,
            url: URLInput,
            log_message: str | list[str] = None,
            **kwargs
    ) -> aiohttp.ClientResponse | None:
        """Handle logging a request, send the request, and return the response"""
        if isinstance(log_message, str):
            log_message = [log_message]
        elif log_message is None:
            log_message = []

        if isinstance(self.session, CachedSession):
            log_message.append("Cached Request")
        self.log(method=method.name, url=url, message=log_message, **kwargs)

        if not isinstance(self.session, CachedSession):
            self._clean_requests_kwargs(kwargs)
        if "headers" in kwargs:
            kwargs["headers"].update(self.session.headers)

        try:
            async with self.session.request(method=method.name, url=url, **kwargs) as response:
                yield response
                if self.wait_timer is not None:
                    await self.wait_timer
        except aiohttp.ClientError as ex:
            self.logger.debug(str(ex))
            yield
            if self.wait_timer is not None:
                await self.wait_timer

    @staticmethod
    def _clean_requests_kwargs(kwargs: dict[str, Any]) -> None:
        """Clean ``kwargs`` by removing any kwarg not in the signature of the :py:meth:`aiohttp.request` method."""
        signature = inspect.signature(aiohttp.request).parameters
        for key in list(kwargs):
            if key not in signature:
                kwargs.pop(key)

    async def _log_response(self, response: aiohttp.ClientResponse, method: HTTPMethod, url: URLInput) -> None:
        """Log the method, URL, response text, and response headers."""
        response_headers = response.headers
        if isinstance(response.headers, Mapping):  # format headers if JSON
            response_headers = json.dumps(dict(response.headers), indent=2)
        self.log(
            method=method.name,
            url=url,
            message=[
                f"Status code: {response.status}",
                "Response text and headers follow:\n"
                f"Response text:\n\t{(await response.text()).replace("\n", "\n\t")}\n"
                f"Headers:\n\t{response_headers.replace("\n", "\n\t")}"
                f"\33[0m"
            ]
        )

    async def _handle_response(self, response: aiohttp.ClientResponse, retry_timer: Timer | None = None) -> bool:
        if response.status not in self.response_handlers:
            return False

        response_handler: StatusHandler = self.response_handlers[response.status]
        return await response_handler(
            response=response,
            authoriser=self.authoriser,
            session=self.session,
            payload_handler=self.payload_handler,
            wait_timer=self.wait_timer,
            retry_timer=retry_timer,
        )

    async def _handle_retry_timer(self, method: HTTPMethod, url: URLInput, timer: Timer | None) -> None:
        if timer is None or not timer.can_increase:
            raise RequestError("Max retries exceeded")

        if not self._retry_logged:
            self.logger.warning(
                f"\33[93mRequest failed. Will retry request {timer.count_remaining} more times "
                f"and timeout in {timer.total_remaining:.2f} seconds...\33[0m"
            )
            self._retry_logged = True

        self.log(
            method=method.name,
            url=url,
            message=f"Request failed: retrying in {int(timer):.2f} seconds..."
        )
        await timer
        timer.increase()

    async def get(self, url: URLInput, **kwargs) -> P:
        """Sends a GET request."""
        kwargs.pop("method", None)
        return await self.request("get", url=url, **kwargs)

    async def post(self, url: URLInput, **kwargs) -> P:
        """Sends a POST request."""
        kwargs.pop("method", None)
        return await self.request("post", url=url, **kwargs)

    async def put(self, url: URLInput, **kwargs) -> P:
        """Sends a PUT request."""
        kwargs.pop("method", None)
        return await self.request("put", url=url, **kwargs)

    async def delete(self, url: URLInput, **kwargs) -> P:
        """Sends a DELETE request."""
        kwargs.pop("method", None)
        return await self.request("delete", url, **kwargs)

    async def options(self, url: URLInput, **kwargs) -> P:
        """Sends an OPTIONS request."""
        kwargs.pop("method", None)
        return await self.request("options", url=url, **kwargs)

    async def head(self, url: URLInput, **kwargs) -> P:
        """Sends a HEAD request."""
        kwargs.pop("method", None)
        kwargs.setdefault("allow_redirects", False)
        return await self.request("head", url=url, **kwargs)

    async def patch(self, url: URLInput, **kwargs) -> P:
        """Sends a PATCH request."""
        kwargs.pop("method", None)
        return await self.request("patch", url=url, **kwargs)

    def __copy__(self):
        """Do not copy handler"""
        return self

    def __deepcopy__(self, _: dict = None):
        """Do not copy handler"""
        return self



"""
Implementations of cache backends which read/write payload responses.
"""
from aiorequestful.cache.backend.base import ResponseCache
from aiorequestful.cache.backend.sqlite import SQLiteCache

CACHE_CLASSES: frozenset[type[ResponseCache]] = frozenset({SQLiteCache})
CACHE_TYPES = frozenset(cls.type for cls in CACHE_CLASSES)



"""
Exceptions relating to response operations.
"""
from aiohttp import ClientResponse

from aiorequestful.exception import AIORequestfulError, HTTPError


class ResponseError(HTTPError):
    """
    Exception raised for errors relating to responses from HTTP requests.

    :param message: Explanation of the error.
    :param response: The :py:class:`ClientResponse` related to the error.
    """
    def __init__(self, message: str = None, response: ClientResponse | None = None):
        self.message = message
        self.response = response
        formatted = f"Status code: {response.status} | {message}" if response else message
        super().__init__(formatted)


class PayloadHandlerError(AIORequestfulError):
    """Error raised when handling a response's payload."""


class StatusHandlerError(AIORequestfulError):
    """Error raised when handling a response based on its status code."""



"""
Implements a ClientSession which is also capable of caching response payload data to a backend.
"""
import contextlib
import logging
from http.client import InvalidURL
from typing import Self, Any, Unpack

from aiohttp import ClientSession, ClientRequest
from aiohttp.payload import JsonPayload

from aiorequestful._utils import format_url_log
from aiorequestful.cache.backend.base import ResponseCache, ResponseRepository
from aiorequestful.cache.response import CachedResponse
from aiorequestful.types import URLInput, RequestKwargs

ClientSession.__init_subclass__ = lambda *_, **__: _  # WORKAROUND: forces disabling of inheritance warning


class CachedSession(ClientSession):
    """
    A modified session which attempts to get/save responses from/to a stored cache before/after sending it.

    :param cache: The cache to use for managing cached responses.
    """

    __slots__ = ("cache",)

    def __init__(self, cache: ResponseCache, **kwargs):
        super().__init__(**kwargs)

        #: The :py:class:`logging.Logger` for this  object
        self.logger: logging.Logger = logging.getLogger(__name__)

        #: The cache to use when attempting to return a cached response.
        self.cache = cache

    async def __aenter__(self) -> Self:
        self.cache = await self.cache.__aenter__()
        await super().__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await super().__aexit__(exc_type, exc_val, exc_tb)
        await self.cache.__aexit__(exc_type, exc_val, exc_tb)

    @contextlib.asynccontextmanager
    async def request(
            self, method: str, url: URLInput, json: Any = None, persist: bool = True, **kwargs: Unpack[RequestKwargs]
    ):
        """
        Perform HTTP request.

        :param method: HTTP request method (such as GET, POST, PUT, etc.)
        :param url: The URL to perform the request on.
        :param json: A JSON serializable Python object to send in the body of the request.
        :param persist: Whether to persist responses returned from sending network requests i.e. non-cached responses.
        :param kwargs: Any other kwargs required for a successful request.
            Arguments passed through to `.aiohttp.ClientSession.request`.
            See aiohttp reference for more info on available kwargs:
            https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientSession.request
        :return: Either the :py:class:`CachedResponse` if a response was found in the cache,
            or the :py:class:`ClientResponse` if the request was sent.
        """
        try:
            url = self._build_url(url)
        except ValueError as exc:
            raise InvalidURL(url) from exc

        kwargs["headers"] = kwargs.get("headers", {}) | dict(self.headers)
        if json is not None:
            kwargs["data"] = JsonPayload(json, dumps=self._json_serialize)

        drop_kwargs = ("allow_redirects",)

        req = ClientRequest(
            method=method,
            url=url,
            loop=self._loop,
            response_class=self._response_class,
            session=self,
            trust_env=self.trust_env,
            **{k: v for k, v in kwargs.items() if k not in drop_kwargs},
        )

        repository = self.cache.get_repository_from_requests(req.request_info)
        response = await self._get_cached_response(req, repository=repository)
        self._log_cache_hit(request=req, response=response)
        if response is None:
            response = await super().request(method=method, url=url, **kwargs)

        yield response

        if persist and repository is not None and response.ok and not isinstance(response, CachedResponse):
            await repository.save_response(response)

    async def _get_cached_response(
            self, request: ClientRequest, repository: ResponseRepository | None
    ) -> CachedResponse | None:
        if repository is None:
            return

        payload = await repository.get_response(request)
        if payload is None:
            return

        if not isinstance(payload, str | bytes):
            repository = self.cache.get_repository_from_url(request.url)
            payload = await repository.serialize(payload)

        return CachedResponse(request=request, payload=payload)

    def _log_cache_hit(self, request: ClientRequest, response: CachedResponse | None) -> None:
        message = "CACHE HIT" if isinstance(response, CachedResponse) else "HTTP REQUEST"
        self.logger.debug(format_url_log(method="CACHE", url=request.url, messages=message))



"""
Resources to handle manipulation of payload data returned by responses into Python objects.
"""
import json
from abc import ABC, abstractmethod
from collections.abc import Awaitable
from typing import Any

from aiohttp import ClientResponse

from aiorequestful.response.exception import PayloadHandlerError
from aiorequestful.types import JSON


class PayloadHandler[T: Any](ABC):
    """Handles payload data conversion to return response payload in expected format."""

    __slots__ = ()

    @abstractmethod
    async def serialize(self, payload: str | bytes | bytearray | T) -> str:
        """Serialize the payload object to a string."""
        raise NotImplementedError

    @abstractmethod
    async def deserialize(self, response: str | bytes | bytearray | ClientResponse | T) -> T:
        """
        Extract payload data from the given ``response`` and serialize to the appropriate object.

        :param response: The response/payload to handle.
        :raise PayloadHandlerError: When the input data is not recognised.
        """
        raise NotImplementedError

    def __call__(self, response: str | bytes | bytearray | ClientResponse | T) -> Awaitable[T]:
        return self.deserialize(response=response)


class JSONPayloadHandler(PayloadHandler[JSON]):

    __slots__ = ("indent",)

    def __init__(self, indent: int = None):
        self.indent = indent

    async def serialize(self, payload: str | bytes | bytearray | JSON) -> str:
        if isinstance(payload, str | bytes | bytearray):
            try:
                payload = json.loads(payload)
            except (json.decoder.JSONDecodeError, TypeError):
                raise PayloadHandlerError(f"Unrecognised input type: {payload}")
        return json.dumps(payload, indent=self.indent)

    async def deserialize(self, response: str | bytes | bytearray | ClientResponse | JSON) -> JSON:
        match response:
            case dict():
                try:  # check the given payload can be converted to/from JSON format
                    return json.loads(json.dumps(response))
                except (json.decoder.JSONDecodeError, TypeError):
                    raise PayloadHandlerError("Given payload is not a valid JSON object")
            case str() | bytes() | bytearray():
                return json.loads(response)
            case ClientResponse():
                return await response.json(content_type=None)
            case _:
                raise PayloadHandlerError(f"Unrecognised input type: {response}")


class StringPayloadHandler(PayloadHandler[str]):

    __slots__ = ()

    async def serialize(self, payload: str | bytes | bytearray) -> str:
        if isinstance(payload, bytes | bytearray):
            return payload.decode()
        return str(payload)

    async def deserialize(self, response: str | bytes | bytearray | ClientResponse) -> str:
        match response:
            case str():
                return response
            case bytes() | bytearray():
                return response.decode()
            case ClientResponse():
                return await response.text()
            case None:
                raise PayloadHandlerError(f"Unrecognised input type: {response}")
            case _:
                return str(response)



"""
Exceptions relating to authoriser operations.
"""
from aiorequestful.exception import AIORequestfulError


class AuthoriserError(AIORequestfulError):
    """Exception raised for errors relating to authorisation."""



"""
Generic utility functions and classes which can be used throughout the entire package.
"""
from collections.abc import Iterator, Iterable
from typing import Any

from aiohttp import RequestInfo
from yarl import URL

from aiorequestful.exception import AIORequestfulImportError
from aiorequestful.types import UnitIterable


class classproperty:
    """Set an immutable class property with this decorator"""
    def __init__(self, func):
        self.fget = func

    def __get__(self, instance, owner):
        return self.fget(owner)


def get_iterator(value: Any) -> Iterator | None:
    """
    Safely get an iterator.

    When a string is given, the iterator will be size 1 where the first element is the given string.
    """
    if value is None:
        return iter(())
    elif isinstance(value, Iterator):
        return value
    elif isinstance(value, str) or isinstance(value, RequestInfo) or not isinstance(value, Iterable):
        value = (value,)
    return iter(value)


def format_url_log(method: str, url: URL, messages: UnitIterable[Any]) -> str:
    """Format a request for a given ``url`` of a given ``method`` appending the given ``messages``"""
    url = str(url.with_query(None))
    url_pad_map = [30, 40, 70, 100]
    url_pad = next((pad for pad in url_pad_map if len(url) < pad), url_pad_map[-1])

    return f"{method.upper():<7}: {url:<{url_pad}} | {" | ".join(map(str, get_iterator(messages)))}"


def required_modules_installed(modules: list, this: object = None) -> bool:
    """Check the required modules are installed, raise :py:class:`AIORequestfulImportError` if not."""
    modules_installed = all(module is not None for module in modules)
    if not modules_installed and this is not None:
        names = [name for name, obj in globals().items() if obj in modules and not name.startswith("_")]
        if isinstance(this, str):
            message = f"Cannot run {this}. Required modules: {", ".join(names)}"
        else:
            message = f"Cannot create {this.__class__.__name__} object. Required modules: {", ".join(names)}"

        raise AIORequestfulImportError(message)

    return modules_installed



"""
Base interface for implementations of all authoriser flows.
"""
import logging
from abc import ABC, abstractmethod
from collections.abc import Awaitable
from typing import Generator, Any

from aiorequestful.types import Headers

_DEFAULT_SERVICE_NAME = "unknown_service"


class Authoriser(ABC):
    """
    Base interface for authenticating and authorising access to a service over HTTP.

    :param service_name: The service name for which to authorise.
    """

    __slots__ = ("logger", "service_name")

    def __init__(self, service_name: str = _DEFAULT_SERVICE_NAME):
        #: The :py:class:`logging.Logger` for this  object
        self.logger: logging.Logger = logging.getLogger(__name__)

        #: The service name for which to authorise. Currently only used for logging purposes.
        self.service_name = service_name

    @abstractmethod
    async def authorise(self) -> Headers:
        """
        Authenticate and authorise, testing/refreshing/re-authorising as needed.

        :raise AuthoriserError: If the authorisation failed to generate valid a token if needed,
            or if the tests continue to fail despite authorising/re-authorising.
        """
        raise NotImplementedError

    def __call__(self) -> Awaitable[Headers]:
        return self.authorise()

    def __await__(self) -> Generator[Any, None, Headers]:
        return self.authorise().__await__()



"""
Exceptions relating to cache operations.
"""
from aiorequestful.exception import AIORequestfulError


class CacheError(AIORequestfulError):
    """Exception raised for errors relating to the cache."""
